@@ -31,7 +31,7 @@ import dataclasses
 import functools
 import pathlib
 import re
-from typing import List, Optional, Any, Set, Type, Union, Generator, Dict
+from typing import Dict, List, Optional, Any, Set, Type, Union, Generator
 
 import jpype
 from jpype._pykeywords import pysafe  # noqa : jpype does not expose a public API for the Java name mangling it applies
@@ -60,6 +60,7 @@ class TypeVarStr:
 class ArgSig:
     name: str
     argType: Optional[TypeStr] = None
+    varArgs: bool = False
 
 
 @dataclasses.dataclass(frozen=True)
@@ -115,36 +116,67 @@ def generateJavaStubs(parentPackages: List[jpype.JPackage],
     Errors in stub generation are treated in a lenient way; failing to generate stubs for one or more java classes
     will not stop stub generation for other classes.
     """
-    packages = []  # type: List[jpype.JPackage]
+    packages: Dict[str, jpype.JPackage] = {}
     for pkg in parentPackages:
-        packages.extend(list(packageAndSubPackages(pkg)))
+        packages.update({pkg.__name__: pkg for pkg in packageAndSubPackages(pkg)})
 
     log.info(f'Collected {len(packages)} packages ...')
 
-    # Map package names to a list of direct subpackages
-    # (e.g {'foo.bar': ['wibble', 'wobble']}).
-    subpackages = collections.defaultdict(list)
-    for pkg in packages:
-        # If this package is a subpackage (i.e. it has a "." in the name) then
-        # get its parent's name, and add the package to the parent's list of
-        # subpackages.
-        if '.' in pkg.__name__:
-            parent, name = pkg.__name__.rsplit('.', 1)
-            subpackages[parent].append(name)
-
+    # Map package names to a set of direct subpackages
+    # (e.g {'foo.bar': {'wibble', 'wobble'}}).
+    subpackages = collections.defaultdict(set)
     outputPath = pathlib.Path(outputDir)
-    for pkg in packages:
-        pathParts = pkg.__name__.split('.')
-        if useStubsSuffix:
-            pathParts[0] += '-stubs'
-        submodulePath = outputPath
-        for pathPart in pathParts:
-            submodulePath = submodulePath / pathPart
-            submodulePath.mkdir(parents=True, exist_ok=True)
-            initFile = submodulePath / '__init__.pyi'
-            initFile.touch()
+    # Prepare a dictionary for *all* package names (including the parents of
+    # the actual packages that we wish to generate stubs for) which maps to the
+    # path of the appropriate __init__.pyi stubfile.
+    stubfilePackagesPaths = {}
+    for pkgName in packages:
+        pkgParts = pkgName.split('.')
 
-        generateStubsForJavaPackage(pkg, submodulePath / '__init__.pyi', subpackages[pkg.__name__], includeJavadoc)
+        submodulePath = outputPath
+        submoduleName = ''
+        for pkgPart in pkgParts:
+            if not submoduleName and useStubsSuffix:
+                submodulePath = submodulePath / f'{pkgPart}-stubs'
+            else:
+                submodulePath = submodulePath / pkgPart
+
+            if not submoduleName:
+                submoduleName = pkgPart
+            else:
+                submoduleName += f'.{pkgPart}'
+
+            if '.' in submoduleName:
+                parent, name = submoduleName.rsplit('.', 1)
+                subpackages[parent].add(name)
+
+            stubfilePackagesPaths[submoduleName] = submodulePath / '__init__.pyi'
+
+    for pkgName, stubfilePath in stubfilePackagesPaths.items():
+        stubfilePath.parent.mkdir(parents=True, exist_ok=True)
+
+        pkg = packages.get(pkgName)
+        if pkg is not None:
+            generateStubsForJavaPackage(pkg, stubfilePath, sorted(subpackages[pkgName]), includeJavadoc)
+        else:
+            importOutput = []
+            classOutput = []
+            generateModuleProtocol(
+                pkgName,
+                [],
+                sorted(subpackages[pkgName]), importOutput, classOutput,
+            )
+            output = []
+
+            for line in sorted(set(importOutput)):
+                output.append(line)
+
+            output.extend([''] * 2)
+            for line in classOutput:
+                output.append(line)
+            with stubfilePath.open('wt') as file:
+                for line in output:
+                    file.write(f'{line}\n')
 
     if jpypeJPackageStubs:
         tld_packages = {name.split('.')[0] for name in subpackages}
@@ -664,14 +696,31 @@ def pythonType(javaType: Any, typeVars: Optional[List[TypeVarStr]] = None, isArg
         return pythonType(jBound, typeVars)
     # Java arrays returned by JPype are of type "JArray", but JArray is not generic. To conserve the type information
     # of the elements, we map them to typing.List for the time being.
-    elif isinstance(javaType, GenericArrayType):
-        return TypeStr('typing.List', [pythonType(javaType.getGenericComponentType(), typeVars)])
-    elif javaType.isArray():
-        return TypeStr('typing.List', [pythonType(javaType.getComponentType(), typeVars)])
+    elif isinstance(javaType, GenericArrayType) or javaType.isArray():
+        elementType = javaArrayComponentType(javaType)
+        listType = TypeStr('typing.List', [pythonType(elementType, typeVars)])
+        if isArgument and str(elementType) == 'byte':
+            # hack: JPype supports converting bytes/bytearray to byte[] but this is not advertised in hints...
+            return TypeStr('typing.Union', [listType, TypeStr('bytes'), TypeStr('bytearray')])
+        else:
+            return listType
     else:
         return translateTypeName(str(javaType.getName()), implicitConversions=isArgument)
 
 
+def javaArrayComponentType(javaType: Any) -> Any:
+    """
+    Get the component type of a java array type (parametrized type for generic arrays, otherwise "standard" type)
+    :param javaType: the array type
+    :return: the component type
+    """
+    from java.lang.reflect import GenericArrayType # noqa
+    if isinstance(javaType, GenericArrayType):
+        return javaType.getGenericComponentType()
+    else:
+        return javaType.getComponentType()
+
+
 def pythonTypeVar(javaType: Any, uniqScopeId: str) -> TypeVarStr:
     """
     Generate python TypeVar definitions for the provided parametrized Java type. This is complicated by the fact that
@@ -856,8 +905,11 @@ def generateJavaMethodStub(parentName: str,
         args = [] if static else [ArgSig(name='self')]  # type: List[ArgSig]
         for jArg in jArgs:
             jArgType = jArg.getParameterizedType()
+            if jArg.isVarArgs():
+                jArgType = javaArrayComponentType(jArgType)
             jArgName = str(jArg.getName()) if jArg.isNamePresent() else inferArgName(jArgType, args)
-            args.append(ArgSig(name=jArgName, argType=pythonType(jArgType, usableTypeVars, isArgument=True)))
+            args.append(ArgSig(name=jArgName, argType=pythonType(jArgType, usableTypeVars, isArgument=True),
+                               varArgs=jArg.isVarArgs()))
 
         signatures.append(JavaFunctionSig(name, args=args, retType=pythonType(jReturnType, usableTypeVars),
                                           static=static, typeVars=methodTypeVars))
@@ -883,6 +935,8 @@ def generateJavaMethodStub(parentName: str,
                 argDef = arg.name
             else:
                 argDef = pysafe(arg.name)
+                if arg.varArgs:
+                    argDef = '*' + argDef
 
                 if arg.argType:
                     argDef += ': ' + toAnnotatedType(arg.argType, parentName, classesDone, classesUsed, importsOutput)
@@ -1081,7 +1135,7 @@ def generateJavaClassStub(package: jpype.JPackage,
     jOverloads = jClass.class_.getMethods()
     for attr, value in items:
         if isinstance(value, jpype.JMethod):
-            matchingOverloads = [o for o in jOverloads if str(o.getName()) == attr and not o.isSynthetic()]
+            matchingOverloads = [o for o in jOverloads if pysafe(str(o.getName())) == attr and not o.isSynthetic()]
             generateJavaMethodStub(packageName, attr, matchingOverloads, javadoc.methods, classesDone=classesDone,
                                    classesUsed=classesUsed, classTypeVars=usableTypeVars, output=methodsOutput,
                                    importsOutput=importsOutput)
